/*

tile2bin - convert graphics to binary data suitable for Dragon/Tandy CoCo
Copyright 2014-2020 Ciaran Anscomb

Licence: GNU GPL version 3 or later <http://www.gnu.org/licenses/gpl-3.0.html>.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.

See the output of "tile2bin --help" for options.

*/

#include <ctype.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef _WIN32
#include <io.h>
#endif

#include <SDL.h>
#include <SDL_image.h>

#ifdef main
# undef main
#endif

static int getpixel(SDL_Surface *surface, int x, int y);
static int getpixel_1bit(SDL_Surface *surface, int x, int y);
static int getpixel_2bit(SDL_Surface *surface, int x, int y);
static int getpixel_4bit(SDL_Surface *surface, int x, int y);

static void helptext(void) {
	puts(
"Usage: tile2bin [OPTION]... IMAGE-FILE\n"
"Convert graphics to binary data suitable for Dragon/Tandy CoCo.\n"
"\n"
"  -x PIXELS     offset to left edge [0]\n"
"  -y PIXELS     offset to top edge [0]\n"
"  -w PIXELS     width [from image]\n"
"  -h PIXELS     height [from image]\n"
"  -X            flip horizontally\n"
"  -Y            flip vertically\n"
"  -s SHIFT      initial right shift in pixels (0-7) [0]\n"
"  -W WIDTH      pixel width in source image [2]\n"
"  -H HEIGHT     pixel height in source image [1]\n"
"  -b BITS       bits per pixel (1, 2, 4) [2]\n"
"  -r            resolution graphics (same as -W 1 -b 1)\n"
"  -m V[,V]...   map output to different values\n"
"  -o FILE       output filename [standard out]\n"
"\n"
"  -?, --help    show this help\n"
	);
}

static struct option long_options[] = {
	{ "help", no_argument, NULL, '?' },
	{ NULL, 0, NULL, 0 }
};

int main(int argc, char **argv) {
	int xoff = 0, yoff = 0;
	int w = 0, h = 0;
	_Bool flip_horizontal = 0;
	_Bool flip_vertical = 0;
	int shift = 0;
	char *mapstring = NULL;
	int pwidth = 2;
	int pheight = 1;
	int bits = 2;
	int map[16] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };

#ifdef _WIN32
	setmode(fileno(stdout), O_BINARY);
#endif

	int c;
	while ((c = getopt_long(argc, argv, "x:y:w:h:XYs:W:H:b:rm:o:?",
				long_options, NULL)) != -1) {
		switch (c) {
		case 0:
			break;
		case 'x':
			xoff = strtol(optarg, NULL, 0);
			break;
		case 'y':
			yoff = strtol(optarg, NULL, 0);
			break;
		case 'w':
			w = strtol(optarg, NULL, 0);
			break;
		case 'h':
			h = strtol(optarg, NULL, 0);
			break;
		case 'X':
			flip_horizontal = 1;
			break;
		case 'Y':
			flip_vertical = 1;
			break;
		case 's':
			shift = strtol(optarg, NULL, 0);
			if (shift < 0 || shift > 7) {
				fprintf(stderr, "shift out of range\n");
				exit(EXIT_FAILURE);
			}
			break;
		case 'W':
			pwidth = strtol(optarg, NULL, 0);
			break;
		case 'H':
			pheight = strtol(optarg, NULL, 0);
			break;
		case 'b':
			bits = strtol(optarg, NULL, 0);
			break;
		case 'r':
			pwidth = bits = 1;
			break;
		case 'm':
			mapstring = strdup(optarg);
			break;
		case 'o':
			if (freopen(optarg, "wb", stdout) == NULL) {
				fprintf(stderr, "Couldn't open '%s' for writing: ", optarg);
				perror(NULL);
				exit(EXIT_FAILURE);
			}
			break;
		case '?':
			helptext();
			exit(EXIT_SUCCESS);
		default:
			exit(EXIT_FAILURE);
		}
	}

	if (optind >= argc) {
		fputs("no input files\n", stderr);
		exit(EXIT_FAILURE);
	}

	if (bits != 1 && bits != 2 && bits != 4) {
		fprintf(stderr, "Invalid number of bits per pixel: use 1, 2 or 4\n");
		exit(EXIT_FAILURE);
	}

	SDL_Surface *in = IMG_Load(argv[optind]);
	if (!in) {
		fprintf(stderr, "Couldn't read image %s: %s\n", argv[optind], IMG_GetError());
		return 1;
	}

	if (mapstring) {
		char *ent = strtok(mapstring, ",");
		for (int i = 0; i < 16; i++) {
			if (ent) {
				map[i] = strtol(ent, NULL, 0);
				ent = strtok(NULL, ",");
			}
		}
	}

	if (!w)
		w = in->w;
	if (!h)
		h = in->h;
	int nlines = h / pheight;
	int wbytes = (((w*bits)/pwidth) + shift + 7) >> 3;  // bytes per line
	int nbytes = wbytes * nlines;  // total bytes

	uint8_t *data = malloc(nbytes);
	uint8_t *datap = data;

	for (int y = 0; (y+pheight) <= h; y += pheight) {
		uint8_t d = 0;
		int n = shift;
		for (int x = 0; x < w; x += pwidth) {
			int getx = flip_horizontal ? (xoff + w - x - 1) : (xoff + x);
			int gety = flip_vertical ? (yoff + h - y - 1) : (yoff + y);
			int c;
			switch (bits) {
			case 1: c = getpixel_1bit(in, getx, gety); break;
			default:
			case 2: c = getpixel_2bit(in, getx, gety); break;
			case 4: c = getpixel_4bit(in, getx, gety); break;
			}
			c = map[c&15];
			d = (d << bits) | c;
			n += bits;
			if (n == 8) {
				n = 0;
				*(datap++) = d;
			}
		}
		if (n) {
			d <<= (8-n);
			*(datap++) = d;
		}
	}

	SDL_FreeSurface(in);
	datap = data;

	// emit h records of wbytes bytes
	fwrite(datap, wbytes, h, stdout);
	return EXIT_SUCCESS;
}

// Orange and black are special-cased, otherwise this maps RGB values to
// encoded VDG colours.  First element, being black, never used.

static int encode_rgb[8] = {
	0, 2, 0, 5, 3, 6, 1, 4
};

// Returns an encoded colour value from a palette of 17:
// 0-7: green, yellow, blue, red, white, cyan, magenta, orange
// 8-15: dark versions of above
// 16: black

static int getpixel(SDL_Surface *surface, int x, int y) {
	if (x < 0 || y < 0 || x >= surface->w || y >= surface->h)
		return -1;
	int bpp = surface->format->BytesPerPixel;
	Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
	Uint32 dp;
	switch(bpp) {
	case 1: dp = *p; break;
	case 2: dp = *(Uint16 *)p; break;
	case 3:
		if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
			dp = p[0] << 16 | p[1] << 8 | p[2];
		else
			dp = p[0] | p[1] << 8 | p[2] << 16;
		break;
	case 4: dp = *(Uint32 *)p; break;
	default: dp = 0; break;
	}
	Uint8 dr, dg, db;
	SDL_GetRGB(dp, surface->format, &dr, &dg, &db);
	// special-case bright orange
	if (dr >= 0xaa && dg >= 0x55 && dg < 0xaa && db < 0x55)
		return 7;
	// special-case dark orange
	if (dr >= 0x55 && dr < 0xaa && dg >= 0x1b && dg < 0x55 && db < 0x55)
		return 15;
	int colour = 0;
	colour |= (dr >= 0x55) ? 0x04 : 0;
	colour |= (dg >= 0x55) ? 0x02 : 0;
	colour |= (db >= 0x55) ? 0x01 : 0;
	// special-case black
	if (colour == 0)
		return 16;
	int encoded = encode_rgb[colour];
	if (dr < 0xaa && dg < 0xaa && db < 0xaa)
		encoded |= 0x08;
	return encoded;
}

static int getpixel_1bit(SDL_Surface *surface, int x, int y) {
	int colour = getpixel(surface, x, y);
	if (colour < 0)
		return -1;
	// return 1 for any "bright" pixel, else 0
	return (colour < 8);
}

static int getpixel_2bit(SDL_Surface *surface, int x, int y) {
	int colour = getpixel(surface, x, y);
	if (colour < 0)
		return -1;
	// works for either PMODE 3 colour set
	return colour & 3;
}

static int getpixel_4bit(SDL_Surface *surface, int x, int y) {
	int colour = getpixel(surface, x, y);
	if (colour < 0)
		return -1;
	// just strips black
	return colour & 15;
}
